/* Voting in Clusters, for finding frequently appearing strings */
/* 2009 Takeaki Uno */

#ifndef _medstr_c_
#define _medstr_c_

#include"vec.c"
#include"seq.c"
#include"sgraph.c"
#include"problem.c"

/* threshold numbers to be given */
int MEDSTR_min_degree = 10;   // minimum degree to be a seed. it would be given by user
int MEDSTR_wildcard_th = -1;   // maximum consecutive-nonsolid letters. it would be given by user
double MEDSTR_dist_factor = 1.5;   // min. dist. between seeds. it would be given by user
double MEDSTR_ovlp_factor = .7;   // min. dist. between cliques
double MEDSTR_unique_char_th = .7;   // min. ratio for a char. to be unique frequent char.
//#define MEDSTR_SIMPAIR_FACTOR 256
unsigned short *MEDSTR_deg = NULL;

/* error messages */
void MEDSTR_error (){
  ERROR_MES = "command explanation";
  print_err ("medstr: comput frequently apperating strings by voting in clusters\n\
usage: medstr iqVcGJdCpNnwP [options] string-filename [string/position/degree-filename/clique-file] [seed-len (if string is not given)] max.mis string.len output-filename\n\
%%:show progress, _:no message, +:write solutions in append mode\n\
i:ignore '\\n', c:multi string comparison\n\
G:genome sequence mode (fasta file), J:(In genome mode), regard '>'(comment)\n\
as a separator, J:(medstr16) read text in Japanese EUC mode\n\
J:(medstr32)regard each byte as 1 digit of at most 200 (32 -> 0,..., 231 -> 199)\n\
0:read number-sequence file\n\
d:read degree file (specify [degree-filename seed-len]),\n\
C:read clique file, p:give position and len. (specify [position seed-len])\n\
n:output numbers instead of letters (valid only in 32-bit mode)\n\
N:output positions of each appearance, S:output positions with string ID\n\
w:use wild card if two or more majorities are there\n\
P:output positions of strings in each cluster, instead of representative strings\n\
  [options]\n\
-b [num]: add [num] to each position/ID (-B [num] is for the second position/ID)\n\
-B [num]: deal with only large degree positions, and output when [num] large degree strings are similar\n\
-F [num]: deal only with positions with degrees no greater than [num]\n\
-t [num]: threshold value to be a majority letter\n\
-w [char]: set the character for the wildcard to [char]\n\
-d [num]: duplicate strings from [num]th to the end\n\
-r [num]: reverse strings from [num]th string to the last string\n\
-X [similarity file]: extend the pattern strings by the clique-similarity\n\
-l,-u [num]: lower/upper bound for outputting strings\n\
-A [file name]: give similarity position file\n\
# the 1st letter of input-filename cannot be '-'.\n");
  EXIT;
}

/* read commands and options from command line */
void MEDSTR_read_param (int argc, char *argv[], PROBLEM *PP){
  int c=1, cc;
  SEQ *SS = &PP->SS;
  PP->th = MEDSTR_unique_char_th;  // default threshold value
  PP->len = 0;
  SS->wildcard = '#';
  PP->plen = 256;

  if ( argc < c+6 ){ MEDSTR_error (); return; }

  cc = c;
  if ( !strchr (argv[c], '_') ) SS->flag |= SHOW_MESSAGE;
  if ( strchr (argv[c], '+') ) PP->II.flag |= ITEMSET_APPEND;
  if ( strchr (argv[c], 'i') ) SS->flag |= SEQ_IGNORE_NEWLINE;
  if ( strchr (argv[c], 'G') ) SS->flag |= SEQ_GENOME;
  if ( strchr (argv[c], 'J') ) SS->flag |= SEQ_EUC;
  if ( strchr (argv[c], '0') ) SS->flag |= SEQ_NUM;
  if ( strchr (argv[c], 'n') ) SS->flag |= SEQ_PUTC_NUM;
  if ( !strchr (argv[c], 'N') && !strchr (argv[c], 'S') ) SS->flag |= SEQ_NO_OUTPUT_POSITION;
  if ( strchr (argv[c], 'S') ) SS->flag |= SEQ_OUTPUT_ID;
  if ( strchr (argv[c], 'c') ) SS->flag |= SEQ_MULTI_STRING;
  if ( strchr (argv[c], 'w') ) PP->problem |= PROBLEM_CLOSED;   // use wildcard
  if ( strchr (argv[c], 'P') ) PP->problem |= PROBLEM_FREQSET;  // clique output
  c++;
  
  while (argv[c][0] == '-'){
    switch (argv[c][1]){
      case 'b': SS->base1 = atoi(argv[c+1]);
      break; case 'B': PP->len = atoi(argv[c+1]);   // deal with only positions with large degrees (valid only in degree file read mode) 
      break; case 't': PP->th = atof(argv[c+1]);
      break; case 's': MEDSTR_wildcard_th = atoi(argv[c+1]);
      break; case 'w': SS->wildcard = argv[c+1][0];
      break; case 'r': SS->reversed_by = atoi(argv[c+1]);
      break; case 'd': SS->duplicated_by = atoi(argv[c+1]);
      break; case 'X': PP->SG.fname = argv[c+1];
      break; case 'l': SS->plen_lb = atoi(argv[c+1]);
      break; case 'u': SS->plen_ub = atoi(argv[c+1]);
      break; case 'A': PP->input_fname = argv[c+1];
      break; case 'F': PP->width = atoi(argv[c+1]);   // upper bound for degree
      break; default: goto NEXT;
    }
    c += 2;
    if ( argc < c+5 ){ MEDSTR_error (); return; }
  }
  
  NEXT:;
  PP->SS.fname = argv[c];
  SS->plen = atoi(argv[c+2]);
  if ( strchr (argv[cc], 'C') ) PP->FF.fname = argv[c+1];
  else if ( strchr (argv[cc], 'd') ) PP->position_fname = argv[c+1];
  else if ( strchr (argv[cc], 'p') ) PP->root = atoi(argv[c+1]);
  else {
    PP->SS.pat = (SEQ_CHAR *)argv[c+1];
    SS->plen = strlen((char *)PP->SS.pat);
    c--;
  }
  SS->perr = atoi(argv[c+3]);
  SS->plen_org = atoi(argv[c+4]);
  PP->output_fname = argv[c+5];
  if ( MEDSTR_wildcard_th < 0 ) MEDSTR_wildcard_th = SS->perr;
}

/* compute the majority string in a cluster, given by positions in jump */
void MEDSTR_vote (PROBLEM *PP, QUEUE *jump, int *cnt, int start, int len, SEQ_CHAR *r){
  SEQ *SS = &PP->SS;
  int l;
  QUEUE_INT *e;
  SEQ_CHAR c;
  SEQ_ID sx;

  FLOOP (l, 0, len) r[l*3] = 0;
  FLOOP (l, 0, len){
    MQUE_FLOOP (*jump, e){
      sx = SS->str_num==2?1: SEQ_get_substID (SS, *e);
      if ( RANGE (SS->head[sx], ((size_t)(*e))+l+start, SS->tail[sx]-1 )){
        cnt[SS->s[((size_t)(*e))+l+start]]++;
      }
    }
    MQUE_FLOOP (*jump, e){
      sx = SS->str_num==2?1: SEQ_get_substID (SS, *e);
      if ( *e+l+start<0 || !RANGE (SS->head[sx], ((size_t)(*e))+l+start, SS->tail[sx]-1)) continue;
      c = SS->s[((size_t)(*e))+l+start];
      if ( ((double)cnt[c])/jump->t >= PP->th ){
//        v[r[l*3]] = ((double)cnt[c])/jump->t;
        r[l*3]++; r[l*3 + r[l*3]] = c;
      }
      cnt[c] = 0;
    }
  }
}

/* print a charactor according to the result of voting */
void MEDSTR_putc (PROBLEM *PP, SEQ_CHAR *r, int l){
  SEQ *SS = &PP->SS;
  if ( r[l*3] == 0 || ((PP->problem&PROBLEM_CLOSED) && r[l*3]>1) ){
#ifdef SEQ_CHAR_LONG
    SEQ_putc (PP->II.fp, SS, 100000000);
#elif defined(SEQ_CHAR_SHORT)
    SEQ_putc (PP->II.fp, SS, 65535);
#else
    fputc (SS->unperm[SS->wildcard], PP->II.fp);
#endif
    if ( SS->flag&SEQ_PUTC_NUM ) fputc (' ', PP->II.fp);
  } else {
    if ( r[l*3]>1 ){
#ifdef SEQ_CHAR_LONG
      SEQ_putc (PP->II.fp, SS, 100000001);
#elif defined(SEQ_CHAR_SHORT)
      SEQ_putc (PP->II.fp, SS, 65534);
#else
      fputc ('[', PP->II.fp);
#endif
      if ( SS->flag&SEQ_PUTC_NUM ) fputc (' ', PP->II.fp);
    }
    SEQ_putc (PP->II.fp, SS, SS->unperm[r[l*3+1]]);
    if ( SS->flag&SEQ_PUTC_NUM ) fputc (' ', PP->II.fp);
    if ( r[l*3]>1 ){
      SEQ_putc (PP->II.fp, SS, SS->unperm[r[l*3+2]]);
#ifdef SEQ_CHAR_LONG
      SEQ_putc (PP->II.fp, SS, 100000002);
#elif defined(SEQ_CHAR_SHORT)
      SEQ_putc (PP->II.fp, SS, 65533);
#else
      fputc (']', PP->II.fp);
#endif
      if ( SS->flag&SEQ_PUTC_NUM ) fputc (' ', PP->II.fp);
    }
  }
}

/* print a string of the result of voting */
void MEDSTR_sim_collect (PROBLEM *PP, SEQ_CHAR *p, int plen, int perr, QUEUE *jump){
  SEQ *SS = &PP->SS;
  QUEUE_INT *v;
  SEQ_LONG j, jj, jjj, q;
  int i, ii, iii, flag = 0;

  jump->t = 0;
  if ( PP->input_fname ){  // case: position similarity file is given
    j = PP->dir; i=0; q = p - PP->SS.s;
//    printf ("%d %d\n", j, PP->OQ[j/MEDSTR_SIMPAIR_FACTOR].t);
    MQUE_FLOOP (PP->OQ[j/PP->plen], v){  // simpair mode
      jjj = *v+j;
      for (ii=PP->itemcand.t; i<ii-1 ; ){
        iii = (i+ii)/2;
        if ( PP->itemcand.v[iii] <= jjj ) i = iii; else ii = iii;
      }
      jj = PP->itemcand.v[i];
//      if ( jj == jjj && MEDSTR_deg[i]>0 && SEQ_Hamming_distance(SS, &SS->s[jj], p, plen, perr) <= perr )
      if ( jj == jjj && SEQ_Hamming_distance(SS, &SS->s[jj], p, plen, perr) <= perr ){
        if ( flag==0 && jj>q ){ flag = 1; QUE_INS (*jump, q); } // ins. pattern string itself
        QUE_INS (*jump, jj);
      }
    }
    if ( flag==0 ) QUE_INS (*jump, q); // pattern string locates at the last
  } else if ( PP->len > 0 ){ // scan all candidents
    MQUE_FLOOP (PP->itemcand, v)
        if ( SEQ_Hamming_distance (SS, &SS->s[*v], p, plen, perr) <= perr ) QUE_INS (*jump, *v);
  } else {
    FLOOP (i, 1, SS->str_num){  // scan all positions
      FLOOP (j, SS->head[i], SS->tail[i]-plen+1){
        if ( SEQ_Hamming_distance (SS, &SS->s[j], p, plen, perr) <= perr ) QUE_INS (*jump, j);
      }
    }
  }
}


/* extend the solid-letter-agreement area */
void MEDSTR_extend_frq (SEQ *SS, SEQ_CHAR *r, int *left, int *right){
  int flag=0, i, f;
  FLOOP (i, SS->plen_org, SS->plen_org+SS->plen) if ( r[i*3] != 1 ) flag++;
  f = flag;
  BLOOP (*left, SS->plen_org, 1){
    if ( r[(*left)*3] != 1 ) f++;
    if ( r[((*left)+SS->plen)*3] != 1 ) f--;
    if ( f > MEDSTR_wildcard_th ) break;
  }
  f = flag;
  FLOOP (*right, SS->plen_org+SS->plen, SS->plen_org*2+SS->plen-1){
    if ( r[(*right)*3] != 1 ) f++;
    if ( r[((*right)-SS->plen)*3] != 1 ) f--;
    if ( f > MEDSTR_wildcard_th ) break;
  }
}

/* compute one frequent string from a string  */
int MEDSTR_comp_frq_str (PROBLEM *PP, SEQ_CHAR *pat, SEQ_CHAR *r, int *left, int *right){
  QUEUE *Q;
  SEQ *SS = &PP->SS;
  SEQ_LONG l;
  QUEUE_INT *x;
  
  if ( PP->FF.fname ) Q = (QUEUE *)pat; 
  else { Q = &PP->vecjump; MEDSTR_sim_collect (PP, pat, SS->plen, SS->perr, Q); }
  if ( Q->t < PP->len ) return (0);
  MEDSTR_vote (PP, Q, PP->itemary, -SS->plen_org, PP->root, r);
  MEDSTR_extend_frq (SS, r, left, right);
  if ( *right-*left < SS->plen_lb || *right-*left > SS->plen_ub ) return (1);

  if ( !(SS->flag & SEQ_NO_OUTPUT_POSITION) ){
    fprintf (PP->II.fp, "%d %d  ", PP->vecjump.t, *right-*left); // output frequency and pattern length
    MQUE_FLOOP (PP->vecjump, x){
      l = *x-SS->plen_org+*left;
      if ( SS->flag & SEQ_OUTPUT_ID ) fprintf (PP->II.fp, "%d:%d ", SEQ_get_substID(SS, *x), (int)(l-SS->head[SEQ_get_substID(SS, *x)]));
      else fprintf (PP->II.fp, "%d ", l);
    }
    fprintf (PP->II.fp, "\n");
  }
  FLOOP (l, *left, *right) MEDSTR_putc (PP, r, l);
  fputc ('\n', PP->II.fp);

    // compute solid substring in frequent string
  return (1);
}

int MEDSTR_degree_cmp (const void *x, const void *y){
 if ( MEDSTR_deg[*((SEQ_LONG *)x)] > MEDSTR_deg[*((SEQ_LONG *)y)] ) return (-1);
 else return (MEDSTR_deg[*((SEQ_LONG *)x)] < MEDSTR_deg[*((SEQ_LONG *)y)]);
}

/* greedy clustering taking highest-degree approach */
void MEDSTR_degree (PROBLEM *PP, SEQ_CHAR *r){
  FILE2 fp;
  SEQ *SS = &PP->SS;
  SEQ_ID sx;
  SEQ_LONG m=0, i, x, y, t, p, j, jj, jjj, xx, xxx, h, *w=NULL;
  int f1, f2;  
  int left, right, perr2 = ((double)SS->perr)*MEDSTR_dist_factor; 
  QUEUE *Q;
  
    // read degree file  (each line is "position,degree" )
  FILE2_open (fp, PP->position_fname, "r", EXIT);
  FLOOP (i, 0, 2){
    t=0;
    while (1){
      x = (SEQ_LONG)FILE2_read_int (&fp);
      y = (SEQ_LONG)FILE2_read_int (&fp);
      if ( FILE_err==6 ) break;
      FILE2_read_until_newline (&fp);
      if ( y < 1 ) continue;   
      if (i==1){ PP->itemcand.v[t] = x; MEDSTR_deg[t] = MIN(y, 65535); }
      t++;
    }
    if ( i==0 ){
      QUEUE_alloc (&PP->itemcand, t+1);
      calloc2 (MEDSTR_deg, t+1, EXIT);
      FILE2_reset (&fp);
    }
  }
  if ( PP->input_fname ){
    calloc2 (w, t+1, EXIT);
    ARY_INIT_PERM (w,t+1);
    qsort (w, t, sizeof(SEQ_LONG), MEDSTR_degree_cmp);
  }
  PP->itemcand.t = t;
  FILE2_close (&fp);
  if ( t==0 ) goto END2;

    // main routine
  h = 0;
  while(1){
//    FLOOP (x, 0, PP->itemcand.t) printf ("%d ", PP->itemcand.v[x]); printf ("%d \n");

      // choose the maximum degree vertex from candidates/itemcand
    if ( PP->input_fname ){
      while ( MEDSTR_deg[w[h]]==0 ) if ( ++h == t ) goto END2;
      i = w[h];
    } else {
      ARY_MAX (m, i, MEDSTR_deg, 0, PP->itemcand.t);
      if ( m==0 ) break;
    }
      // show progress
    if ( PP->prog2 < PP->prog*100/t )
        { PP->prog2 = PP->prog*100/t; print_mes (SS, "%d/100 (%d)\n",PP->prog2, MEDSTR_deg[i]); }

    PP->dir = p = PP->itemcand.v[i];
    if ( MEDSTR_deg[i] > PP->width || !MEDSTR_comp_frq_str (PP, &SS->s[p], r, &left, &right) ){ MEDSTR_deg[i] = 0; PP->prog++; continue; } 
//    FLOOP (x, 0, PP->root){ SS->pat[x] = r[x*3+1]; printf ("%c", SS->unperm[SS->pat[x]]); } printf ("\n");
    FLOOP (x, 0, PP->root) SS->pat[x] = r[x*3+1];
      // clear parts of solid-frequent strings

    if ( PP->input_fname ){
      Q = &PP->OQ[p/PP->plen];
//      printf ("%d %d\n", Q->t, PP->itemcand.t);
      y=0 ; FLOOP (x, 0, PP->itemcand.t){  // simpair mode
        do { // find pair (x,y) st. itemcand.v[x]-plen_org <= Q->v[y] <= itemcand.v[x]+plen_org
          j = MAX (SS->plen_org,PP->itemcand.v[x])-SS->plen_org;
          while ( (jj=Q->v[y]+p) < j ) if ( ++y == Q->t ) goto END;
          jjj = MAX (jj,SS->plen_org) - SS->plen_org;
          for (xx=PP->itemcand.t ; x<xx-1 ; ){
            xxx = (x+xx)/2;
            if (PP->itemcand.v[xxx] < jjj) x = xxx; else xx = xxx;
          }
          j = PP->itemcand.v[x];
        } while ( jj < MAX (j,SS->plen_org) -SS->plen_org );
        if ( MEDSTR_deg[x]>0 ){
          for ( xx=y; xx<Q->t && (m=(Q->v[xx]+p))<j+SS->plen_org ; xx++ ){
            if ( (f1=SEQ_Hamming_distance (SS, &SS->s[j], &SS->pat[SS->plen_org+(j-m)], SS->plen, perr2)) <= perr2 ){
              sx = SS->str_num==2?1: SEQ_get_substID (SS, m);
              if ( RANGE (SS->head[sx], m, SS->tail[sx]-SS->plen )){
                if ( (f2=SEQ_Hamming_distance (SS, &SS->s[m], &SS->pat[SS->plen_org], SS->plen, perr2)) <= perr2 ){ MEDSTR_deg[x] = 0; PP->prog++; goto NEXT2; }
              }
            }
          }
        }
        NEXT2:;
      }
      END:;
          // clear the degree of positions near by i (current operating position)
      for (j=i ; j>0 && PP->itemcand.v[j]+SS->plen_org >= p ; )
          { j--; if (  MEDSTR_deg[j] > 0 ) PP->prog++; MEDSTR_deg[j] = 0; }
      for (j=i ; j<PP->itemcand.t && PP->itemcand.v[j] <= p+SS->plen_org ; j++ )
          { if (  MEDSTR_deg[j] > 0 ) PP->prog++; MEDSTR_deg[j] = 0; }
    } else {    // scan all candidents
      FLOOP (x, 0, PP->itemcand.t){
        if ( MEDSTR_deg[x]==0 || x==i ) { MEDSTR_deg[x] = 0; goto NEXT; }
        FLOOP (y, MAX(left-SS->perr,0), MIN(SS->plen_org*2, right-SS->plen+SS->perr)){
          if ( (f1=SEQ_Hamming_distance (SS, &SS->s[PP->itemcand.v[x]], SS->pat+y, SS->plen, perr2)) <= perr2 ){
            if ( (f2=SEQ_Hamming_distance (SS, &SS->s[PP->itemcand.v[x] +(SS->plen_org-y)],
                 SS->pat+SS->plen_org, SS->plen, perr2)) <= perr2 ){ MEDSTR_deg[x]=0; PP->prog++; goto NEXT; }
          }
        }
        NEXT:;
      }
    }

    if ( !PP->input_fname ){
      p=0; FLOOP (x, 0, PP->itemcand.t)
        if ( MEDSTR_deg[x]>0 ){ MEDSTR_deg[p] = MEDSTR_deg[x]; PP->itemcand.v[p++] = PP->itemcand.v[x]; }
      PP->itemcand.t = p;
    }
  }
  END2:;
  free2 (MEDSTR_deg);
  free2 (w);
}

/* greedy clustering taking highest-degree approach */
/*
void MEDSTR_clique (PROBLEM *PP, SEQ_CHAR *r){
  SEQ *SS = &PP->SS;
  QUEUE_INT *v, *vv;
  SEQ_LONG m, i;
  QUEUE_ID x;
  int left, right, shift, th, *cnt;
  
  calloc2 (cnt, PP->root, "MEDSTR_clq: cnt", EXIT);
  while(1){
    ST_MAX (m, i, PP->FF.v, t, 0, PP->FF.t);
    if ( m==0 ) break;

    MEDSTR_comp_frq_str (PP, (SEQ_CHAR *)(&PP->FF.v[i]), r, &left, &right);
    FLOOP (x, 0, PP->root) SS->pat[x] = r[x*3+1];

      // remove overlapping cliques to the current clique
    FLOOP (x, 0, PP->FF.t){
      if ( PP->FF.v[x].t == 0 || x==i ) continue;  // jump to end if the clique is already cleared
      th = ((double)MIN (PP->FF.v[x].t, PP->FF.v[i].t)) * MEDSTR_ovlp_factor;
      
         // count the overlap, for each shift
      m=0; MQUE_FLOOP (PP->FF.v[x], v){
        MQUE_FLOOP (PP->FF.v[i], vv){
          shift = *v - *vv;
          if ( RANGE (left, shift+SS->plen_org, right-SS->plen) ){
            cnt[shift + SS->plen_org] ++;
            ENMAX (m, cnt[shift + SS->plen_org]);
            if ( m >= th ) goto NEXT;
          }
        }
      }
      NEXT:;
         // clear the count, used above
      MQUE_FLOOP (PP->FF.v[x], v){
        MQUE_FLOOP (PP->FF.v[i], vv){
          shift = *v - *vv;
          if ( RANGE (left, shift+SS->plen_org, right-SS->plen) ) cnt[shift + SS->plen_org] =0;
        }
      }
         // if overlap is no less than th, clear the clique
      if ( m >= th ) PP->FF.v[x].t = 0;
    }
      // clear the queue of maximum degree
    PP->FF.v[i].t = 0;
  }
  free2 (cnt);
}
*/


/***************************************************************************/
/***************************************************************************/
SEQ_LONG MEDSTR_read_simpair (PROBLEM *PP){
  SEQ *SS = &PP->SS;
  FILE2 fp;
  size_t tt;
  SEQ_LONG x, y, z, t;
  int i, j, flag;
  QUEUE_INT *v;

  FILE2_open (fp, PP->input_fname, "r", EXIT);
  fseek (fp.fp, 0, SEEK_END);
  tt = ftell (fp.fp);
  ENMIN (PP->plen, ((LONG)SS->str_len)*256L/((LONG)tt));
  ENMAX (PP->plen, 16);
  fseek (fp.fp, 0, SEEK_SET);

  malloc2 (PP->OQ, SS->str_len/PP->plen+1, EXIT);
  ARY_FILL (PP->OQ, 0, SS->str_len/PP->plen+1, INIT_QUEUE);

  FLOOP (i, 0, 2){
    t=0;
    while (1){
      x = (SEQ_LONG)FILE2_read_int (&fp);
      y = (SEQ_LONG)FILE2_read_int (&fp);
      z = (SEQ_LONG)FILE2_read_int (&fp);
      if ( FILE_err==6 ) break;
      FILE2_read_until_newline (&fp);
      if ( y < 1 ) continue;
      FLOOP (j, x/PP->plen, (x+z-PP->SS.plen)/PP->plen+1){
        if ( i==0 ){ t++; PP->OQ[j].t++; }
        else {
          flag=0; MQUE_FLOOP (PP->OQ[j], v) if ( *v==y-x ){ flag=1; break; }
          if ( !flag ) QUE_INS (PP->OQ[j], y-x);
        }
      }
      FLOOP (j, y/PP->plen, (y+z-PP->SS.plen)/PP->plen+1){
        if ( i==0 ){ t++; PP->OQ[j].t++; }
        else {
          flag=0; MQUE_FLOOP (PP->OQ[j], v) if ( *v==x-y ){ flag=1; break; }
          if ( !flag ) QUE_INS (PP->OQ[j], x-y);
        }
      }
    }
    if ( i==0 ){
      if ( t==0 ){ FILE2_close (&fp); return (1); }
      calloc2 (PP->OQ->v, t+1, EXIT);
      FLOOP (j, 1, SS->str_len/PP->plen+1){
        PP->OQ[j].v = PP->OQ[j-1].v + PP->OQ[j-1].t;
        PP->OQ[j-1].t = 0;
      }
      PP->OQ[j-1].t = 0;
      FILE2_reset (&fp);
    }
  }
  FILE2_close (&fp);
  FLOOP (j, 0, SS->str_len/PP->plen+1)
      qsort_int (PP->OQ[j].v, PP->OQ[j].t, 1);
  return (0);
}

/* initialization for main routine */
void MEDSTR_init (PROBLEM *PP){
  SEQ *SS = &PP->SS;
  double l;
  int i;

//  PP->SG.flag = LOAD_ARC;
  PROBLEM_load (PP);  // load the sequence file
  if ( PP->SS.pat ) FLOOP (i, 0, SS->plen) SS->pat[i] = SS->perm[SS->pat[i]];
  else SS->pat = &(SS->s[PP->root]);
  if ( PP->width == 0 ) PP->width = SS->str_len;  // no upper bound for degrees
  
    // read similarity file
  if (PP->input_fname){
    if ( MEDSTR_read_simpair (PP) ){ ERROR_MES = "no pair is given"; puts (ERROR_MES); }
  }

    // open output file
//  if (PP->output_fname) fopen2 (PP->II.fp, PP->output_fname, "w", "MEDSTR:output_file", EXIT);

  l = MAX (SS->str_len/32, MIN (SS->str_len, 1000000));
  PROBLEM_alloc (PP, MIN (SEQ_CHAR_END, l), l, 0, NULL, PROBLEM_VECJUMP + PROBLEM_ITEMARY + PROBLEM_VECMARK);
  PP->root = SS->plen+SS->plen_org*2;
}

/*******************************************************************/
int MEDSTR_main (int argc, char *argv[]){
  PROBLEM PP;
  SEQ *SS = &PP.SS;
  SEQ_CHAR *r;
  int left, right;
  
  PROBLEM_init (&PP);
  MEDSTR_read_param (argc, argv, &PP);
if ( ERROR_MES ) return (1);
  MEDSTR_init (&PP);
  calloc2 (r, PP.root*3, EXIT0);

  if ( !ERROR_MES ){
    if ( !PP.position_fname && !PP.FF.fname)
        MEDSTR_comp_frq_str (&PP, PP.SS.pat, r, &left, &right);
    else {
      calloc2 (SS->pat, PP.root, EXIT);
      if ( PP.position_fname ) MEDSTR_degree (&PP, r);
//      else MEDSTR_clique (&PP, r);
      free2 (SS->pat);
    }
  }
  SS->pat=NULL;
  PROBLEM_end (&PP);
  free2 (r);
  return (ERROR_MES?1:0);
}

/*******************************************************************************/
#ifndef _NO_MAIN_
#define _NO_MAIN_
int main (int argc, char *argv[]){
  return (MEDSTR_main (argc, argv));
}
#endif
/*******************************************************************************/


#endif

