/* SACHICA: Scalable Algorithm for Charactaristic and Homologineous Interval CAluclation */
/* 2004 Takeaki Uno */
// #include"stdlib2.c"

#ifndef _sachica_h_
#define _sachica_h_

#include"pos.h"
#include"problem.h"

/* un-comment the following definition if post-process for each pair is needed.
 for each pair found, foo(SACHICA *SA, SEQ_ID sx, SEQ_ID sy, SEQ_LONG x, SEQ_LONG y, int core_id)
 is called, where x and y are positions of pair, sx and sy are string ID of positions,
 SA is the SACHICA structure including all data SACHICA is using, and core_id is the core_id
 of the CPU core which operates the process, in the multi-core mode (in normal mode, always 1).
 The function foo can be renamed to any name. */

// #define SACHICA_POSTPROCESS foo 

// BMP HEADER
#pragma pack(1)
typedef struct _BMPFILEHEADER {
  unsigned short bfType;
  unsigned int  bfSize;
  unsigned short bfReserved1;
  unsigned short bfReserved2;
  unsigned int  bfOffBits;
} BMPFILEHEADER;

typedef struct _BMPINFOHEADER{
    unsigned int  biSize;
    int           biWidth;
    int           biHeight;
    unsigned short biPlanes;
    unsigned short biBitCount;
    unsigned int  biCompression;
    unsigned int  biSizeImage;
    int           biXPixPerMeter;
    int           biYPixPerMeter;
    unsigned int  biClrUsed;
    unsigned int  biClrImportant;
} BMPINFOHEADER;
#pragma pack()


#define SACHICA_PHASE 1       // flag for phase 1 or phase 2
#define SACHICA_MISMATCH 2    // flag for computing mismatch torellance
#define SACHICA_PAIR     4    // flag for finding all similar pairs
#define SACHICA_REPEAT   8    // flag for finding all repeating pairs
#define SACHICA_PRINT    16    // flag for printing given substring of input string
#define SACHICA_REVERSE_X 32   // reverse file for x direction
#define SACHICA_REVERSE_Y 64   // reverse file for y direction
 
#define SACHICA_INTERLEAVE 128 // select positions to be compared with interleave
#define SACHICA_MINIMAL 256    // output only canonical position-pairs
#define SACHICA_DOUBLE_READ 512 // each two fragments are considered to be a pair
#define SACHICA_EXT_VALID 1024  // extract maximum valid strings including few ignoring char's
#define SACHICA_OVERLAP1  2048  // output string pair only if one is overlapping with the other
#define SACHICA_OVERLAP 4096  // output string pair only if one is included in the other
#define SACHICA_ONLY_MINIMUM 8192 // output a pair only when it is shortest distance for at least one of them
#define SACHICA_LENGTH_HEAD 32768   // output interval length by #substrings in it

#define SACHICA_SOLEXA 65536    // solexa analyze mode
#define SACHICA_ASSEMBLY 131072 // print corresponding substrings to each position
#define SACHICA_MASK_STRING 262144 // print corresponding substrings to each position
#define SACHICA_E_CLUSTERING 524288    // epsiron-clustering mode
#define SACHICA_NO_MULTI 1048576    // do not multi-string compare
#define SACHICA_ONE_SIMILAR 2097152    // output similar string if the ratio of the length of similar strings is no less than that of at least one string

#define SACHICAM_DRAW_GRID  65536 // draw frid lines
#define SACHICAM_DETECT  131072 // detect homologeneous substrings
#define SACHICAM_HOMOLOGY_REVERSE_X 262144  // homology string detection, reverse X direction
#define SACHICAM_HOMOLOGY_REVERSE_Y 524288  // homology string detection, reverse X direction
#define SACHICAM_LINE    1048276 // write long line representing points having similar one
#define SACHICAM_LINE_SKIP 5   // interval of parallel lines in line mode
#define SACHICAM_MIRROR    2097152 // write long line representing points having similar one
#define SACHICAM_BMP    4194304 // write long line representing points having similar one
#define SACHICAM_LONG_SKIP 8388608  // allow long skip without dots, for diagonal line detection
#define SACHICAM_READ_INTERVAL 16777216  // read interval format file for line mode
#define SACHICAM_NEGATIVE 33554432  // highlight small intesity dots, for line mode
#define SACHICAM_POSITION_SC 67108864  // read the file as position-score file

#define SACHICA_DOUBLE_READ_GAP 4000
#define SOLEXAASM_BITMASK 32
#define SOLEXAASM_SC_MAX_RATIO 1.5
#define SOLEXAASM_SC_LB_RATIO 0.7

#ifndef SACHICA_CELL_CHAR
  #define SACHICA_CELL_CHAR unsigned short
  #define SACHICA_CELL_END 65536
#endif

typedef struct {
  SEQ_ID *ID;
  int IDend;
//  unsigned short *totail, *tohead;   // buffers used in pairwise comparison
  int *totail, *tohead;   // buffers used in pairwise comparison
  char *flag_a, *flag_b;  // flags for each position used in pairwis
  int core_id;
} SACHICA_MUL;

typedef struct {
  PROBLEM PP;
  SACHICA_MUL *MUL;

  int gap_ub, gap_lb;   // upper/lower bounds for gaps
  int th, th1, th2;   // thresholds
  
  SEQ_LONG *headmax, *tailmax;  // for masking; boundary of valid strings 
  SEQ_LONG count1, count2;
  int plen;
  
  SEQ_ID **in, **out, *in_t, *out_t, *buf, *buf2, *cnt;
  int *mark;
  SEQ_LONG *x;
  QUEUE cand, jump, occ;
  SEQ_ID str_num;        // preserving original #strings 
  char *in_tc, *out_tc;  // for small degrees of each vertex
  int gap_max, gap_min;  // maximum gap and minimum gap

  int *str_sc;
  SEQ_LONG *lbuf;               // buffer for SEQ_LONG
  char *cbuf, *cbuf2;          // buffer for CHAR
  FILE *fp, *fp1, *fp2;
  
  // for picture drawing
  SEQ_LONG xmax, ymax, xmax2, ymax2;   // maximum x/y-coordinates among the pairs
  int pxmax, pymax, pxmax2, pymax2;    // picture size
  SACHICA_CELL_CHAR **cell;   // array for cell
  int boxx, boxy;   // size of each cell: length of string
  int color;
  double density;   // color number and density
  int line_skip, col_num;
  int emphasize;    // emphasize each dot to rectangle of 2*this width/height
  
  int pw_flag1, pw_flag2, pw_flag3, pw_flag4; // flags used in pairwise comparison
  int iter2_factor;  // threshold factor for proceeding to iter2 or not
  SEQ_LONG trancate;  // upper bound of group size for doing pairwise comparison

  // for sachicax
//  SEQ_LONG *slot1, *slot2, *slot, *slotx1, *slotx2, *flag1, *flag2;
//  size_t slot1_end, slot2_end, slot3_end;
} SACHICA;

int SACHICA_col1[16] = {2, 0, 2, 0, 2, 0, 2, 1, 2, 2, 0, 0, 1, 2, 1, 2 };
int SACHICA_col2[16] = {2, 2, 0, 0, 2, 2, 0, 2, 1, 2, 1, 2, 0, 0, 2, 1 };
int SACHICA_col3[16] = {2, 2, 2, 2, 0, 0, 0, 2, 2, 1, 2, 1, 2, 1, 0, 0 };

typedef struct {
#ifdef MULTI_CORE
  pthread_t thr; // thread identifier
#endif
  SACHICA *SA;
  POS *PS;
  SEQ_LONG *pos;
  SEQ_LONG t;
  int block;
  int mis;
  int dep;
  int block_end;
  int level;
  int branch_flag;
  SEQ_LONG pr;
  int core_id;
  int *lock_i;
  SEQ_LONG *lock_s;
  int *lock_prog;
} SACHICA_MULTI_CORE;


// to find similar pairs of short substrings, use p command
//  - if output is huge, use -k 200
//  - for multi string comparison, use -c 2
//  - for genome comparison, duplicate second string, and use -c 2, -r 3

// find short string having large mismatch torelance, use msM,   
// find short string having many similar strings, use msM, -k 250,  
// find short string having few similar strings, use ms, -k 250,  
// find repeat strings, use rsMI, -k [num] 
// find unique strings, use rsI, -k [num] 

// for masking, concat all fragment, plus 2 of mask string. separate here. reverse the last, 
// for solexa, concat all fragment, plus 2 of database string. separate here. reverse the last, 

#ifdef SACHICA_POSTPROCESS
void SACHICA_postprocess (SACHICA *SA, SEQ_ID sx, SEQ_LONG x, SEQ_ID sy, SEQ_LONG y, int k, int core_id);
void SACHICA_postprocess_end (SACHICA *SA);
void SACHICA_postprocess_init (SACHICA *SA);

#endif

/* extract valid substrings from input strings */
#define SACHICA_VALID_TH 2
void SACHICA_extract_valid_string ( SACHICA *SA );

/* check whether a part of x is a substring of the mask string */
void SACHICA_mask_check (SACHICA *SA, SEQ_LONG x, SEQ_LONG y );
void SACHICA_mask_pos (SACHICA *SA );
/* output masked strings */
void SACHICA_output_masked_string (SACHICA *SA);

/* compute #mismatches in the remaining blocks
  if the mismatches > SS->perr, return SS->plen
  if the current combination of the blocks is not canonical for the given pair, return S->plen */
//int SACHICA_remain_mismatch ( SACHICA *SA, SEQ_LONG x, SEQ_LONG y, int block, int block_end, QUEUE *remain_block );
//int SACHICA_remain_mismatch_ ( SACHICA *SA, SEQ_LONG x, SEQ_LONG y, int block, int block_end, QUEUE *remain_block );

/* classify pattern strings beginning at positions in pos_list */
/* for positions having the same pattern strings, set their sc to 0 */
void SACHICA_classify (SEQ *S, SEQ_LONG *pos, SEQ_LONG t, int dep);

/* remove duplicated pattern strings (those sc[]=0) from position list */
void SACHICA_reduce_dup (SEQ *S);

void SACHICA_error ();

/* read commands and options from command line */
void SACHICA_read_param (int argc, char *argv[], SACHICA *SA);

/* temporary memory string ID's, length to head/tail etc., for pairwise comparison */
void SACHICA_set_cache (SACHICA *SA, SACHICA_MUL *MUL, SEQ_LONG *pos, SEQ_LONG t);

/* pairwise compare of all pattern strings in position list  */
void SACHICA_pairwise_comparison (SACHICA *SA, POS *PS, SACHICA_MUL *M, SEQ_LONG *pos, SEQ_LONG t, int block, int block_end);

/* iteration of multi-classification algorithm */
void SACHICA_iter (SACHICA *SA, POS *PS, SACHICA_MUL *MUL, SEQ_LONG *pos, SEQ_LONG t, int block, int mis, int dep, int level, int block_end, int branch_flag);
//void SACHICA_iter ( SACHICA *SA, SEQ_LONG *pos, SEQ_LONG t, int block, int mis, int dep, QUEUE *jump, SEQ_LONG **cnt, QUEUE *remain_block, int core_id );
void *SACHICA_iter_multi (void *p);

/* initialization for main routine */
void SACHICA_init (SACHICA *SA);

#endif


