/************************************************************************/
/* simstr.c: find frequently appearing string patterns from a string, by 
   collecting frequently appearing short substrings and extend it unless the 
   "voting" does not succeed (the mojority letter in a position of them is
   not determined) */
/* 17/Mar/2010 Takeaki Uno  */

#ifndef _simstr_c_
#define _simstr_c_

#ifdef _NO_MAIN_
#define _simstr_c_no_main_
#else
#define _NO_MAIN_
#endif

#define STDLIB2_RADIX_SORT    // use fast integer radix sort, written in stdlib2.c

#include<string.h>
#include"vec.c"
#include"sgraph.c"
#include"sachica.c"
#include"medstr.c"


int SIMSTR_flag=0, SIMSTR_len=0, SIMSTR_plen, SIMSTR_mis, SIMSTR_stop_mis=-1, SIMSTR_ext=0;
int SIMSTR_mes=SHOW_MESSAGE, SIMSTR_frq=2, SIMSTR_no_remove=0;
int SIMSTR_len_ub=0, SIMSTR_len_lb=0, SIMSTR_repeat_factor=5, SIMSTR_repeat_ub=9;
char *SIMSTR_infname, *SIMSTR_outfname, *SIMSTR_workdir = "";
double SIMSTR_letter_th = 0.5;

/* error routine */
void SIMSTR_error (){
  ERROR_MES = "command explanation";
  print_err ("simstr: tGiwN [options] input-file substring-length #mismatch output-file\n\
%%:show progress, _:no message, +:write solutions in append mode, =:do not delete temporal files\n\
t:ascii text comparison mode, G:genome sequence mode, J:(In genome mode), regard '>'\n\
(comment) as a separator, J:read text in Japanese EUC mode (sachica16)\n\
J:(32bit mode)regard each byte as 1 digit of at most 200 (32 -> 0,..., 231 -> 199).\n\
0:read number-sequence file\n\
c:multi-string mode (not compare substrings in the same string)\n\
C:multi-string mode (compare substrings in the same string)\n\
i:ignore newline code, w:use wildcard charactor in the output\n\
n:output numbers instead of letters (valid only in 32-bit mode)\n\
N:output appearing positions of each pattern string, S:output positions with string ID\n\
I:do not use interleave position sampling for finding similar strings\n\
\n[options]\n\
-e [num]:check the similarity for [num] letters in both directions (default=[substring-length * 100])\n\
-f [num]:specify the lower bound of the similar substrings for each substring to be a candidate of seed (default=1)\n\
-F [num]: deal only with positions with degrees no greater than [num]\n\
-s [num]:stop the extension of string when [num] wildcards appear in a string of length [substring-length]\n\
-l,-u [num]:specify the lower/upper bound of output pattern strings\n\
-r [num]:specify the factor of repeat string (shift of [len]/[num] letters with at most [len]/[num] mismatches will be considered as repeat string\n\
-t [num]:specify the threshold of majority letter (default=0.5)\n\
-W [dir]:specify the working directory (folder). The last letter of the directory has to be '/' ('\\')\n\
The 1st letter of input-filename has to be '-'\n");
  EXIT;
}

/* read parameters given by command line */
void SIMSTR_read_param (int argc, char *argv[], char *comm, char *comm2){
  int c=1;

  if ( argc<c+5 ){ SIMSTR_error (); return; }
  comm[0] = 0; comm2[0] = 0;
  strcat (comm, "ZM");
  strcat (comm2, "d");
  if ( !strchr (argv[1], 'I') ) strcat (comm, "l");
  if ( strchr (argv[1], '+') ) strcat (comm2, "+");
  if ( strchr (argv[1], '_') ){ SIMSTR_mes = 0; strcat (comm, "_"); strcat (comm2, "_"); }
  if ( strchr (argv[1], '%') ){ strcat (comm, "%"); strcat (comm2, "%"); }
  if ( strchr (argv[c], '=') ) SIMSTR_no_remove = 1;
  if ( strchr (argv[1], 'i') ){ strcat (comm, "i"); strcat (comm2, "i"); }
  if ( strchr (argv[1], 'c') ){ strcat (comm, "c"); strcat (comm2, "c"); }
  if ( strchr (argv[1], 'C') ){ strcat (comm, "C"); strcat (comm2, "c"); }
  if ( strchr (argv[1], 'J') ){ strcat (comm, "J"); strcat (comm2, "J"); }
  if ( strchr (argv[1], '0') ){ strcat (comm, "0"); strcat (comm2, "0"); }
  if ( strchr (argv[1], 'w') ){ strcat (comm2, "w"); }
  if ( strchr (argv[1], 'N') ){ strcat (comm2, "N"); }
  if ( strchr (argv[1], 'n') ){ strcat (comm2, "n"); }
  if ( strchr (argv[1], 'S') ){ strcat (comm2, "S"); }
  if ( strchr (argv[1], 'G') ){ strcat (comm, "G -S N"); strcat (comm2, "G"); }
  c++;

  while ( argv[c][0] == '-' ){
    if ( argc<c+5 ){ SIMSTR_error (); return; }
    switch ( argv[c][1] ){
      case 'e': if ( (SIMSTR_ext = atoi(argv[c+1])) < 0 )
          error_num("the extend length has to be non-negative",atoi(argv[c+1]), EXIT);
      break; case 'f': if ( (SIMSTR_frq = atoi(argv[c+1])) <= 0 )
          error_num("degree lower bound has to be positive",atoi(argv[c+1]), EXIT);
      break; case 'r': if ( (SIMSTR_repeat_factor = atoi(argv[c+1])) <= 0 )
          error_num("repeat string factor has to be positive",atoi(argv[c+1]), EXIT);
          SIMSTR_repeat_ub = INTHUGE;
      break; case 'M': strcat (comm, " -M "); strcat (comm, argv[c+1]);
      break; case 'l': strcat (comm2, " -l "); strcat (comm2, argv[c+1]);
      break; case 'u': strcat (comm2, " -u "); strcat (comm2, argv[c+1]);
      break; case 't': strcat (comm2, " -t "); strcat (comm2, argv[c+1]);
      break; case 's': strcat (comm2, " -s "); strcat (comm2, argv[c+1]);
      break; case 'F': strcat (comm2, " -F "); strcat (comm2, argv[c+1]);
      break; case 'W': SIMSTR_workdir = argv[c+1];
    }
    c += 2;
  }

  SIMSTR_infname = argv[c];    // input file name
  SIMSTR_plen = atoi(argv[c+1]);  // length
  SIMSTR_mis = atoi(argv[c+2]);  // #mismatches
  if ( SIMSTR_stop_mis < 0 ) SIMSTR_stop_mis = SIMSTR_mis;
  SIMSTR_outfname = argv[c+3];  // output file name
  if ( SIMSTR_ext == 0 ) SIMSTR_ext = SIMSTR_plen * 100;
}

/******************************************************************************/
int SIMSTR_main (int argc, char *argv[]){
  int flag=1;
  char sachica_comm[200], medstr_comm[200], *W = SIMSTR_workdir;

    // read params
  SIMSTR_read_param (argc, argv, sachica_comm, medstr_comm);
  if ( ERROR_MES ) return (1);

///////////////////////////////////////////////////////////
  EXECSUB (SACHICA_main, SIMSTR_mes, goto END, "sachica r%s%s -R %d %d -k %d -l %d \"%s\" %d %d \"%s__tmp__\"", 
      sachica_comm, strchr (sachica_comm, 'G')? " -o -1 -d 1 -r 2": "",
      SIMSTR_plen/SIMSTR_repeat_factor, MIN (SIMSTR_repeat_ub, SIMSTR_plen/SIMSTR_repeat_factor), 
//      SIMSTR_frq, SIMSTR_plen*(strchr (sachica_comm, 'l')? 2: 1), SIMSTR_infname,
      SIMSTR_frq, SIMSTR_plen*(strchr (sachica_comm, 'l')? 1: 1), SIMSTR_infname,
      SIMSTR_plen, SIMSTR_mis, W);

  EXECSUB (SACHICA_main, SIMSTR_mes, goto END, "sachica p%s%s -R %d %d -l %d \"%s\" %d %d \"%s__tmp2__\"", 
      sachica_comm, strchr (sachica_comm, 'G')? " -o -1 -d 1 -r 2": "",
      SIMSTR_plen/SIMSTR_repeat_factor, MIN (SIMSTR_repeat_ub, SIMSTR_plen/SIMSTR_repeat_factor),
      SIMSTR_plen*(strchr (sachica_comm, 'l')? 1: 1), SIMSTR_infname, SIMSTR_plen, SIMSTR_mis, W);
//      SIMSTR_plen*(strchr (sachica_comm, 'l')? 2: 1), SIMSTR_infname, SIMSTR_plen, SIMSTR_mis, W);

  EXECSUB (MEDSTR_main, SIMSTR_mes, goto END, "medstr %s%s -B %d -A \"%s__tmp2__\" \"%s\" \"%s__tmp__\" %d %d %d \"%s\"", 
      medstr_comm, strchr (sachica_comm, 'G')? " -d 1 -r 2": "",
      SIMSTR_frq, W, SIMSTR_infname, W, SIMSTR_plen, SIMSTR_stop_mis, SIMSTR_ext, SIMSTR_outfname);
  flag = 0;

///////////////////////////////////////////////////////////
  END:;
  if ( !SIMSTR_no_remove ) MREMOV (W, "__tmp__", "__tmp2__");
  return (flag);
}

/*******************************************************************************/
#ifndef _simstr_c_no_main_
int main (int argc, char *argv[]){
  return (SIMSTR_main (argc, argv));
}
#endif
/*******************************************************************************/

#endif




